from compas.geometry import Vector
from compas.geometry import Cone
from compas.geometry import Circle
from compas.geometry import Cylinder
from compas.geometry import Plane
from compas.geometry import Shape
from compas.datastructures import Mesh


class Arrow(Shape):
    """A Arrow is defined by its location and direction vector.

    Parameters
    ----------
    position : list of float or `compas.geometry.Vector`
        The start position of arrow
    direction : list of float or `compas.geometry.Vector`
        The height of the cone.

    Attributes
    ----------
    head_portion : float
        The portion of head of the arrow
    head_width : float
        The head width relative to the length of arrow
    body_width : float
        The body width relative to the length of arrow


    Examples
    --------
    >>> from compas.geometry import Arrow
    >>> arrow = Arrow([0, 0, 0], [0, 0, 1])

    """

    def __init__(
            self, position=[0, 0, 0],
            direction=[0, 0, 1],
            head_portion=0.3, head_width=0.1, body_width=0.02):
        super().__init__()
        self.position = Vector(*position)
        self.direction = Vector(*direction)
        self.head_portion = head_portion
        self.head_width = head_width
        self.body_width = body_width

    @property
    def data(self):
        """Returns the data dictionary that represents the Arrow.

        Returns
        -------
        dict
            The arrow data.

        """
        return {'position': list(self.position), 'direction': list(self.direction)}

    @data.setter
    def data(self, data):
        self.position = Vector(*data['position'])
        self.direction = Vector(*data['direction'])

    # ==========================================================================
    # customisation
    # ==========================================================================

    def __repr__(self):
        return 'Arrow({0}, {1})'.format(self.position, self.direction)

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_data(cls, data):
        """Construct a Arrow from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        Arrow
            The constructed arrow.

        Examples
        --------
        >>> from compas.geometry import Arrow
        >>> from compas.geometry import Vector
        >>> data = {'position': Vector(0, 0, 0), 'direction': Vector(0, 0, 1)}
        >>> arrow = Arrow.from_data(data)

        """
        arrow = cls()
        arrow.data = data
        return arrow

    # ==========================================================================
    # methods
    # ==========================================================================

    def to_vertices_and_faces(self, u=4):
        """Returns a list of vertices and faces.

        Parameters
        ----------
        u : int, optional
            Number of faces in the "u" direction.
            Default is ``4``.

        Returns
        -------
        (vertices, faces)
            A list of vertex locations and a list of faces,
            with each face defined as a list of indices into the list of vertices.
        """
        if u < 3:
            raise ValueError('The value for u should be u > 3.')

        # Construct the head of arrow
        head_position = self.position + self.direction*(1-self.head_portion)
        plane = Plane(head_position, self.direction)
        circle = Circle(plane, self.head_width*self.direction.length)
        cone = Cone(circle, self.direction.length*self.head_portion)
        v, f = cone.to_vertices_and_faces(u=u)
        head = Mesh.from_vertices_and_faces(v, f)

        # COnstruct the body of arrow
        body_center = self.position + self.direction*(1-self.head_portion)/2
        plane = Plane(body_center, self.direction)
        circle = Circle(plane, self.body_width*self.direction.length)
        cylinder = Cylinder(circle, self.direction.length * (1 - self.head_portion))
        v, f = cylinder.to_vertices_and_faces(u=u)
        body = Mesh.from_vertices_and_faces(v, f)

        body.join(head)

        return body.to_vertices_and_faces()

    def transform(self, transformation):
        """Transform the Arrow.

        Parameters
        ----------
        transformation : :class:`Transformation`
            The transformation used to transform the cone.

        """
        self.position.transform(transformation)
        self.direction.transform(transformation)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    import doctest
    doctest.testmod()



from .worker import Worker  # noqa : F401
from .timer import Timer  # noqa : F401
from .selector import Selector  # noqa : F401

from .controller import Controller  # noqa : F401
from .app import App  # noqa : F401



from .dockform import DockForm
from qtpy import QtWidgets
from qtpy import QtGui
from qtpy import QtCore
import ast


class ValueDelegate(QtWidgets.QStyledItemDelegate):

    def get_value(self, index):
        treeform = self.parent()
        item = treeform.tree.itemFromIndex(index)
        column = index.column()
        key = treeform.column_keys[column]
        value = item.entry[key]
        return value

    def set_value(self, index, new_value):
        treeform = self.parent()
        item = treeform.tree.itemFromIndex(index)
        column = index.column()
        key = treeform.column_keys[column]
        value = item.entry[key]
        if self.has_options(value):
            value["value"] = new_value
        else:
            item.entry[key] = new_value

    def has_options(self, value):
        return isinstance(value, dict) and "options" in value

    def createEditor(self, parent, option, index):
        value = self.get_value(index)
        if self.has_options(value):
            editor = QtWidgets.QComboBox(parent)
            editor.addItems([str(o) for o in value["options"]])
            return editor
        else:
            editor = QtWidgets.QLineEdit(parent)
            return editor

    def setEditorData(self, editor, index):
        value = self.get_value(index)
        if self.has_options(value):
            editor.setCurrentIndex(value["options"].index(value["value"]))
        else:
            editor.setText(str(value))

    def setModelData(self, editor, model, index):
        value = self.get_value(index)
        if self.has_options(value):
            new_value = value["options"][editor.currentIndex()]
        else:
            new_value = editor.text()
        self.set_value(index, new_value)
        model.setData(index, new_value, QtCore.Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class TreeForm(DockForm):
    def __init__(self, app, title="Tree", data=None, columns=["key", "value"], show_headers=True, striped_rows=False):
        super().__init__(app, title)
        self.column_names = list(map(lambda c: c if isinstance(c, str) else c["name"], columns))
        self.column_keys = list(map(lambda c: c if isinstance(c, str) else c["key"], columns))
        self.column_editable = list(map(lambda c: c.get("editable") if isinstance(c, dict) else False, columns))
        self.striped_rows = striped_rows
        self.tree = QtWidgets.QTreeWidget()
        self.tree.setColumnCount(len(columns))
        self.tree.setHeaderLabels(self.column_names)
        self.tree.setHeaderHidden(not show_headers)
        self.tree.setItemDelegateForColumn(1, ValueDelegate(self))
        self.setWidget(self.tree)
        self.tree.itemDoubleClicked.connect(self.on_item_double_clicked)
        self.tree.itemPressed.connect(self.on_item_pressed)
        self.tree.itemChanged.connect(self.on_item_changed)
        self.tree.currentItemChanged.connect(self.on_current_item_changed)
        self.datastore = []
        if data:
            self.update(data)

    def update(self, entries):
        self.tree.clear()
        self.item_count = 0
        self.map_entries(entries)
        for item in self.items:
            if item.entry.get("expanded"):
                item.setExpanded(True)
        self.tree.resizeColumnToContents(0)

    @property
    def items(self):
        items = []

        def traverseNode(item):
            for i in range(item.childCount()):
                items.append(item.child(i))
                traverseNode(item.child(i))

        for i in range(self.tree.topLevelItemCount()):
            items.append(self.tree.topLevelItem(i))
            traverseNode(self.tree.topLevelItem(i))

        return items

    @property
    def entries(self):
        return list(map(lambda i: i.entry, self.items))

    def map_entries(self, entries, parent=None):

        # automatically convert dict to list of key, value pairs
        if isinstance(entries, dict):
            _entries = []
            for key, value in entries.items():
                _entries.append({"key": key, "value": value})
            entries = _entries

        for entry in entries:
            children = entry.get("children")
            values = []
            colors = []
            for key in self.column_keys:
                value = entry.get(key, "")
                color = None
                if isinstance(value, dict):
                    if "color" in value:
                        color = value["color"]
                    if "value" in value:
                        value = value["value"]
                values.append(value)
                colors.append(color)
            item = QtWidgets.QTreeWidgetItem([str(v) for v in values])
            item.entry = entry
            item.data_item = {"values": values, "children": []}
            if children:
                self.map_entries(children, item)
            if entry.get("color"):
                for i in range(self.tree.columnCount()):
                    item.setBackgroundColor(i, QtGui.QColor(*entry.get("color")))
            elif self.striped_rows and self.item_count % 2 == 1:
                for i in range(self.tree.columnCount()):
                    item.setBackgroundColor(i, QtGui.QColor(240, 240, 240))

            if entry.get("on_item_double_clicked"):
                item.on_item_double_clicked = entry.get("on_item_double_clicked")

            if entry.get("on_item_pressed"):
                item.on_item_pressed = entry.get("on_item_pressed")

            if entry.get("on_item_edited"):
                item.on_item_edited = entry.get("on_item_edited")

            for i, color in enumerate(colors):
                if color:
                    item.setBackgroundColor(i, QtGui.QColor(*color))

            self.item_count += 1

            if parent is not None:
                parent.addChild(item)
                parent.data_item["children"].append(item.data_item)
            else:
                self.tree.addTopLevelItem(item)
                self.datastore.append(item.data_item)

    def on_item_double_clicked(self, item, column):
        if self.column_editable[column]:
            self.tree.openPersistentEditor(item, column)
        if hasattr(item, "on_item_double_clicked"):
            item.on_item_double_clicked(self, item.entry)

    def on_item_pressed(self, item, column):
        if hasattr(item, "on_item_pressed"):
            item.on_item_pressed(self, item.entry)

    def on_item_changed(self, item, column):
        self.tree.closePersistentEditor(item, column)
        try:
            value = ast.literal_eval(item.text(column))
        except Exception:
            value = item.text(column)

        item.data_item["values"][column] = value
        if hasattr(item, "on_item_edited"):
            item.on_item_edited(self, item.entry, column, value)

    def on_current_item_changed(self, current_item, privious_item):
        if privious_item:
            for i in range(self.tree.columnCount()):
                self.tree.closePersistentEditor(privious_item, i)

    def select(self, entries):
        all_items = self.tree.findItems("", QtCore.Qt.MatchContains | QtCore.Qt.MatchRecursive)
        for item in all_items:
            if item.entry in entries:
                item.setSelected(True)
            else:
                item.setSelected(False)



from compas.datastructures import Network
from compas.datastructures import Mesh
from compas.geometry import Vector
from compas.geometry import Point
from compas.geometry import Pointcloud
from compas.geometry import Line
from compas.geometry import Polyline
from compas.geometry import Frame
from compas.geometry import Box
from compas.geometry import Sphere
from compas.geometry import Torus
from compas.geometry import Cylinder
from compas.geometry import Plane
from compas.geometry import Circle
from compas.geometry import Ellipse
from compas.geometry import Polygon
from compas.geometry import Cone
from compas.geometry import Capsule
from compas.geometry import Polyhedron
from compas.geometry import NurbsSurface
from compas.robots import RobotModel

from compas_view2.shapes import Arrow
from compas_view2.shapes import Text
from compas_view2.collections import Collection

from .object import Object
from .bufferobject import BufferObject  # noqa : F401

from .vectorobject import VectorObject
from .pointobject import PointObject
from .pointcloudobject import PointcloudObject
from .lineobject import LineObject
from .polylineobject import PolylineObject
from .frameobject import FrameObject

from .networkobject import NetworkObject
from .meshobject import MeshObject

from .boxobject import BoxObject
from .sphereobject import SphereObject
from .torusobject import TorusObject
from .arrowobject import ArrowObject
from .textobject import TextObject
from .collectionobject import CollectionObject
from .gridobject import GridObject  # noqa : F401
from .cylinderobject import CylinderObject
from .planeobject import PlaneObject
from .circleobject import CircleObject
from .ellipseobject import EllipseObject
from .polygonobject import PolygonObject
from .coneobject import ConeObject
from .capsuleobject import CapsuleObject
from .polyhedronobject import PolyhedronObject

from .nurbssurfaceobject import NurbsSurfaceObject
from .robotobject import RobotObject

try:
    from compas_occ.brep import BRep
    from .brepobject import BRepObject
except ImportError:
    BRep = None
    BRepObject = None

from .object import DATA_OBJECT  # noqa : F401

Object.register(Point, PointObject)
Object.register(Vector, VectorObject)
Object.register(Plane, PlaneObject)
Object.register(Frame, FrameObject)

Object.register(Pointcloud, PointcloudObject)

Object.register(Line, LineObject)
Object.register(Polyline, PolylineObject)

Object.register(Circle, CircleObject)
Object.register(Ellipse, EllipseObject)
Object.register(Polygon, PolygonObject)

Object.register(Box, BoxObject)
Object.register(Sphere, SphereObject)
Object.register(Torus, TorusObject)
Object.register(Cylinder, CylinderObject)
Object.register(Cone, ConeObject)
Object.register(Capsule, CapsuleObject)
Object.register(Polyhedron, PolyhedronObject)

Object.register(Arrow, ArrowObject)
Object.register(Text, TextObject)

Object.register(NurbsSurface, NurbsSurfaceObject)

Object.register(Network, NetworkObject)
Object.register(Mesh, MeshObject)

Object.register(Collection, CollectionObject)
Object.register(RobotModel, RobotObject)

if BRep and BRepObject:
    Object.register(BRep, BRepObject)



from OpenGL import GL
# from PIL import Image

import os
import numpy as np

from compas.geometry import transform_points_numpy

from compas_view2.objects import BufferObject
from compas_view2.objects import TextObject
from compas_view2.objects import VectorObject
from compas_view2.shaders import Shader

from .view import View


class View120(View):
    """View widget for OpenGL version 2.1 and GLSL 120 with a Compatibility Profile.
    """

    def init(self):
        self.grid.init()
        # init the buffers
        for guid in self.objects:
            obj = self.objects[guid]
            obj.init()

        projection = self.camera.projection(self.app.width, self.app.height)
        viewworld = self.camera.viewworld()
        transform = np.identity(4)
        # create the program
        self.shader_model = Shader(name='120/model')
        self.shader_model.bind()
        self.shader_model.uniform4x4("projection", projection)
        self.shader_model.uniform4x4("viewworld", viewworld)
        self.shader_model.uniform4x4("transform", transform)
        self.shader_model.uniform1i("is_selected", 0)
        self.shader_model.uniform1f("opacity", self.opacity)
        self.shader_model.uniform3f("selection_color", self.selection_color)
        self.shader_model.release()

        self.shader_text = Shader(name='120/text')
        self.shader_text.bind()
        self.shader_text.uniform4x4("projection", projection)
        self.shader_text.uniform4x4("viewworld", viewworld)
        self.shader_text.uniform4x4("transform", transform)
        self.shader_text.uniform1f("opacity", self.opacity)
        self.shader_text.release()

        self.shader_arrow = Shader(name='120/arrow')
        self.shader_arrow.bind()
        self.shader_arrow.uniform4x4("projection", projection)
        self.shader_arrow.uniform4x4("viewworld", viewworld)
        self.shader_arrow.uniform4x4("transform", transform)
        self.shader_arrow.uniform1f("opacity", self.opacity)
        self.shader_arrow.uniform1f("aspect", self.app.width / self.app.height)
        self.shader_arrow.release()

        self.shader_instance = Shader(name='120/instance')
        self.shader_instance.bind()
        self.shader_instance.uniform4x4("projection", projection)
        self.shader_instance.uniform4x4("viewworld", viewworld)
        self.shader_instance.uniform4x4("transform", transform)
        self.shader_instance.release()

        self.shader_grid = Shader(name='120/grid')
        self.shader_grid.bind()
        self.shader_grid.uniform4x4("projection", projection)
        self.shader_grid.uniform4x4("viewworld", viewworld)
        self.shader_grid.uniform4x4("transform", transform)
        self.shader_grid.release()

    def update_projection(self, w=None, h=None):
        w = w or self.app.width
        h = h or self.app.height

        projection = self.camera.projection(w, h)
        self.shader_model.bind()
        self.shader_model.uniform4x4("projection", projection)
        self.shader_model.release()

        self.shader_text.bind()
        self.shader_text.uniform4x4("projection", projection)
        self.shader_text.release()

        self.shader_arrow.bind()
        self.shader_arrow.uniform4x4("projection", projection)
        self.shader_arrow.uniform1f("aspect", w / h)
        self.shader_arrow.release()

        self.shader_instance.bind()
        self.shader_instance.uniform4x4("projection", projection)
        self.shader_instance.release()

        self.shader_grid.bind()
        self.shader_grid.uniform4x4("projection", projection)
        self.shader_grid.release()

    def resize(self, w, h):
        self.update_projection(w, h)

    def sort_objects_from_viewworld(self, viewworld):
        """Sort objects by the distances from their bounding box centers to camera location"""
        opaque_objects = []
        transparent_objects = []
        centers = []
        for guid in self.objects:
            obj = self.objects[guid]
            if isinstance(obj, BufferObject):
                if obj.opacity * self.opacity < 1 and obj.bounding_box_center is not None:
                    transparent_objects.append(obj)
                    centers.append(transform_points_numpy([obj.bounding_box_center], obj.matrix)[0])
                else:
                    opaque_objects.append(obj)
        if transparent_objects:
            centers = transform_points_numpy(centers, viewworld)
            transparent_objects = sorted(zip(transparent_objects, centers), key=lambda pair: pair[1][2])
            transparent_objects, _ = (zip(*transparent_objects))
        return opaque_objects + list(transparent_objects)

    def paint(self):
        viewworld = self.camera.viewworld()
        if self.current != self.PERSPECTIVE:
            self.update_projection()

        # Draw instance maps
        if self.app.selector.enabled:
            self.shader_instance.bind()
            # set projection matrix
            self.shader_instance.uniform4x4("viewworld", viewworld)
            if self.app.selector.select_from == "pixel":
                self.app.selector.instance_map = self.paint_instances()
            if self.app.selector.select_from == "box":
                self.app.selector.instance_map = self.paint_instances(self.app.selector.box_select_coords)
            self.app.selector.enabled = False
            self.clear()
            self.shader_instance.release()

        # Draw grid
        self.shader_grid.bind()
        self.shader_grid.uniform4x4("viewworld", viewworld)
        if self.app.selector.wait_for_selection_on_plane:
            self.app.selector.uv_plane_map = self.paint_plane()
            self.clear()
        if self.show_grid:
            self.grid.draw(self.shader_grid)
        self.shader_grid.release()

        # Draw model objects in the scene
        self.shader_model.bind()
        self.shader_model.uniform4x4("viewworld", viewworld)
        for obj in self.sort_objects_from_viewworld(viewworld):
            if obj.is_visible:
                obj.draw(self.shader_model, self.mode == "wireframe", self.mode == "lighted")
        self.shader_model.release()

        # draw arrow sprites
        self.shader_arrow.bind()
        self.shader_arrow.uniform4x4("viewworld", viewworld)
        for guid in self.objects:
            obj = self.objects[guid]
            if isinstance(obj, VectorObject):
                if obj.is_visible:
                    obj.draw(self.shader_arrow)
        self.shader_arrow.release()

        # draw text sprites
        self.shader_text.bind()
        self.shader_text.uniform4x4("viewworld", viewworld)
        for guid in self.objects:
            obj = self.objects[guid]
            if isinstance(obj, TextObject):
                if obj.is_visible:
                    obj.draw(self.shader_text)
        self.shader_text.release()

        # draw 2D box for multi-selection
        if self.app.selector.select_from == "box":
            self.shader_model.draw_2d_box(self.app.selector.box_select_coords, self.app.width, self.app.height)

        if self.app.record:
            # r = self.devicePixelRatio()
            # x, y, width, height = 0, 0, self.app.width, self.app.height
            # buffer = GL.glReadPixels(x*r, y*r, width*r, height*r, GL.GL_RGB, GL.GL_UNSIGNED_BYTE)
            # arr = np.frombuffer(buffer, dtype=np.uint8).reshape(height*r, width*r, 3)
            # arr = arr[::-r, ::r, :]
            # im = Image.fromarray(arr, mode="RGB")
            # im = im.convert(mode='RGB', dither=0)
            # print("recording frame:", self.app.frame_count)
            # self.app.recorded_frames.append(im)
            qimage = self.grabFramebuffer()
            qimage.save(os.path.join(self.app.tempdir, f"{self.app.frame_count}.png"), "png")

    def paint_instances(self, cropped_box=None):
        GL.glDisable(GL.GL_POINT_SMOOTH)
        GL.glDisable(GL.GL_LINE_SMOOTH)
        if cropped_box is None:
            x, y, width, height = 0, 0, self.app.width, self.app.height
        else:
            x1, y1, x2, y2 = cropped_box
            x, y = min(x1, x2), self.app.height - max(y1, y2)
            width, height = abs(x1 - x2), abs(y1 - y2)
        for guid in self.objects:
            obj = self.objects[guid]
            if hasattr(obj, "draw_instance"):
                if obj.is_visible:
                    obj.draw_instance(self.shader_instance, self.mode == "wireframe")
        # create map
        r = self.devicePixelRatio()
        instance_buffer = GL.glReadPixels(x*r, y*r, width*r, height*r, GL.GL_RGB, GL.GL_UNSIGNED_BYTE)
        instance_map = np.frombuffer(instance_buffer, dtype=np.uint8).reshape(height*r, width*r, 3)
        instance_map = instance_map[::-r, ::r, :]
        GL.glEnable(GL.GL_POINT_SMOOTH)
        GL.glEnable(GL.GL_LINE_SMOOTH)
        return instance_map

    def paint_plane(self):
        x, y, width, height = 0, 0, self.app.width, self.app.height
        self.grid.draw_plane(self.shader_grid)
        r = self.devicePixelRatio()
        plane_uv_map = GL.glReadPixels(x*r, y*r, width*r, height*r, GL.GL_RGB, GL.GL_FLOAT)
        plane_uv_map = plane_uv_map.reshape(height*r, width*r, 3)
        plane_uv_map = plane_uv_map[::-r, ::r, :]
        return plane_uv_map



from qtpy import QtWidgets

from compas_view2.objects import DATA_OBJECT

from .form import Form


class AddForm(Form):
    def __init__(self, on_create=None):
        super().__init__('Add Object')

        for data_cls in DATA_OBJECT:
            self.add_create_button(data_cls, on_create)

    def add_create_button(self, data_cls, on_create):
        object_cls = DATA_OBJECT[data_cls]
        if hasattr(object_cls, 'create_default'):
            layout = QtWidgets.QHBoxLayout()
            self._inputs.addLayout(layout)
            button = QtWidgets.QPushButton()
            button.setText(data_cls.__name__)

            def _on_create():
                default_data = object_cls.create_default()
                on_create(default_data)
                self.accept()
            button.clicked.connect(_on_create)
            layout.addWidget(button)

    def inputs(self):
        """Creates layout for input fields

        Returns
        -------
        QtWidgets.QVBoxLayout
        """
        return QtWidgets.QVBoxLayout()



from compas_view2.gl import make_index_buffer
from compas_view2.gl import make_vertex_buffer

from .object import Object


class VectorObject(Object):
    """Object for displaying vector as arrow."""

    def __init__(self, data, color=None, position=None, size=1, **kwargs):
        super().__init__(data, **kwargs)
        self.color = color or [0, 0, 0]
        self.position = position or [0, 0, 0]
        self.size = size

    def init(self):
        self.make_buffers()
        self._update_matrix()

    def make_buffers(self):
        self._vector_buffer = {
            'positions': make_vertex_buffer(list(self.position)),
            'directions': make_vertex_buffer(list(self._data)),
            'colors': make_vertex_buffer(self.color),
            'sizes': make_vertex_buffer([self.size]),
            'elements': make_index_buffer([0]),
            'n': 1
        }

    def draw(self, shader):
        shader.enable_attribute('position')
        shader.enable_attribute('direction')
        shader.enable_attribute('color')
        shader.enable_attribute('size')
        shader.uniform4x4('transform', self.matrix)
        shader.bind_attribute('position', self._vector_buffer['positions'])
        shader.bind_attribute('direction', self._vector_buffer['directions'])
        shader.bind_attribute('color', self._vector_buffer['colors'])
        shader.bind_attribute('size', self._vector_buffer['sizes'], step=1)
        shader.draw_arrows(elements=self._vector_buffer['elements'], n=self._vector_buffer['n'])
        shader.disable_attribute('position')
        shader.disable_attribute('direction')
        shader.disable_attribute('color')
        shader.disable_attribute('size')



from .form import Form  # noqa: F401
from .propertyform import PropertyForm  # noqa: F401
from .addform import AddForm  # noqa: F401
# from .point import PointForm  # noqa: F401
# from .line import LineForm  # noqa: F401
# from .sphere import SphereForm  # noqa: F401
# from .torus import TorusForm  # noqa: F401
# from .pointedit import PointEditForm  # noqa: F401



from compas.utilities import flatten

from compas_view2.gl import make_index_buffer
from compas_view2.gl import make_vertex_buffer

from .bufferobject import BufferObject


class GridObject(BufferObject):
    """Object for displaying a grid of lines in the XY plane of the world coordinate system."""

    default_color_lines = [0.75, 0.75, 0.75]

    def __init__(self, cell_size, x_cells, y_cells, background=True):
        super().__init__({}, name="Grid", show_lines=True)
        self.cell_size = cell_size
        self.x_cells = x_cells
        self.y_cells = y_cells
        self.background = background

    def _lines_data(self):
        positions = []
        colors = []
        elements = []
        color = self.default_color_lines
        n = 0
        for x in range(- self.x_cells, self.x_cells + 1):
            if x == 0:
                positions.append([x * self.cell_size, -self.x_cells * self.cell_size, 0])
                positions.append([x * self.cell_size, 0, 0])
                colors.append(color)
                colors.append(color)
                positions.append([x * self.cell_size, 0, 0])
                positions.append([x * self.cell_size, self.x_cells * self.cell_size, 0])
                colors.append([0, 1, 0])
                colors.append([0, 1, 0])
                n = len(elements)*2
                elements.append([n + 0, n + 1])
                elements.append([n + 2, n + 3])
            else:
                positions.append([x * self.cell_size, -self.x_cells * self.cell_size, 0])
                positions.append([x * self.cell_size, self.x_cells * self.cell_size, 0])
                colors.append(color)
                colors.append(color)
                n = len(elements)*2
                elements.append([n, n + 1])

        for y in range(- self.y_cells, self.y_cells + 1):
            if y == 0:
                positions.append([-self.y_cells * self.cell_size, y * self.cell_size, 0])
                positions.append([0, y * self.cell_size, 0])
                colors.append(color)
                colors.append(color)
                positions.append([0, y * self.cell_size, 0])
                positions.append([self.y_cells * self.cell_size, y * self.cell_size, 0])
                colors.append([1, 0, 0])
                colors.append([1, 0, 0])
                n = len(elements)*2
                elements.append([n + 0, n + 1])
                elements.append([n + 2, n + 3])
            else:
                positions.append([-self.y_cells * self.cell_size, y * self.cell_size, 0])
                positions.append([self.y_cells * self.cell_size, y * self.cell_size, 0])
                colors.append(color)
                colors.append(color)
                n = len(elements)*2
                elements.append([n, n + 1])
        return positions, colors, elements

    def init(self):
        self.make_buffers()

        # Create uv plane
        x_size = self.x_cells * self.cell_size
        y_size = self.y_cells * self.cell_size
        positions = [[-x_size, -y_size, 0], [x_size, -y_size, 0], [x_size, y_size, 0], [-x_size, y_size, 0]]
        color = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]]
        elements = [[0, 1, 3], [1, 2, 3], [1, 0, 3], [2, 1, 3]]

        self._uvplane = {
            'positions': make_vertex_buffer(list(flatten(positions))),
            'colors': make_vertex_buffer(list(flatten(color))),
            'elements': make_index_buffer(list(flatten(elements))),
            'n': len(list(flatten(elements))),
        }

    def draw(self, shader):
        """Draw the object from its buffers"""
        shader.enable_attribute('position')
        shader.enable_attribute('color')
        shader.bind_attribute('position', self._lines_buffer['positions'])
        shader.bind_attribute('color', self._lines_buffer['colors'])
        shader.draw_lines(width=self.linewidth, elements=self._lines_buffer['elements'], n=self._lines_buffer['n'], background=True)
        shader.disable_attribute('position')
        shader.disable_attribute('color')

    def draw_plane(self, shader):
        shader.enable_attribute('position')
        shader.enable_attribute('color')
        shader.bind_attribute('position', self._uvplane['positions'])
        shader.bind_attribute('color', self._uvplane['colors'])
        shader.draw_triangles(elements=self._uvplane['elements'], n=self._uvplane['n'])
        shader.disable_attribute('position')
        shader.disable_attribute('color')



from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
import matplotlib
matplotlib.use('Qt5Agg')


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, width=5, height=4, dpi=100):
        self.figure = Figure(figsize=(width, height), dpi=dpi)
        self.figure.update = self.draw
        super(MplCanvas, self).__init__(self.figure)



from .bufferobject import BufferObject


class PointcloudObject(BufferObject):
    """Object for displaying COMPAS point geometry.

    Parameters
    ----------
    data : :class: `compas.datastructures.Pointcloud`
        Pointcloud to be displayed.
    name : string
        Name of the object.
    is_selected : bool
        Whether the object is selected.
    color : list
        Float rgb color for point cloud.
    colors : list
        list of rgb colors for each points, length must equal to number of points.
        If provided, the `color` parameter will be ignored.
    size : float
        The point size to be drawn on screen.

    Attributes
    ----------
    colors : list
        list of point colors

    Raises
    -------
    ValueError
        If number of colors does not equal to number of points.

    """

    def __init__(self, data, **kwargs):
        super().__init__(data, **kwargs)
        self.show_points = True

    def _points_data(self):
        positions = self._data
        colors = [self.pointcolors.get(i, self.pointcolor) for i in range(len(positions))]
        elements = [[i] for i in range(len(positions))]
        return positions, colors, elements



from qtpy import QtCore


class Timer:

    def __init__(self, interval, callback, singleshot=False):
        self.timer = QtCore.QTimer()
        self.timer.setInterval(interval)
        self.timer.timeout.connect(callback)
        self.timer.setSingleShot(singleshot)
        self.timer.start()

    def stop(self):
        self.timer.stop()



from qtpy import QtCore
from qtpy import QtWidgets


class Slider:
    """Class representing a horizontal slider wrapped in a grid layout with two rows.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The app containing the widget.
    parent : QtWidget
        The parent widget.
    action : callable
        The callback connected to the slide action.
    value : int, optional
        The initial value of the slider.
    minval : int, optional
        The minimum value of the sliding range.
    maxval : int, optional
        The maximum value of the sliding range.
    step : int, optional
        Size of value increments.
    title : str, optional
        Title label.
    annotation : str, optional
        Value annotation label.
    interval : int, optional
        The tick interval size.
    bgcolor : Color, optional
        Background color of the box containing the slider.

    Attributes
    ----------
    action : callable
        Action associated with the button click event.
    slider : QtWidgets.QSlider
        The actual slider widget.
    value : float
        The current value of the slider.

    Class Attributes
    ----------------
    STYLE : str
        Stylesheet for the visual appearance of the groove and handle of the slider.

    """

    STYLE = """
QSlider::groove::horizontal {
    border: 1px solid #cccccc;
    background-color: #eeeeee;
    height: 4px;
}
QSlider::handle:horizontal {
    background-color: #ffffff;
    border: 1px solid #cccccc;
    border-radius: 6px;
    height: 12px;
    width: 12px;
    margin: -6px 0;
}
"""

    def __init__(self,
                 app,
                 parent,
                 action,
                 value=0,
                 minval=0,
                 maxval=100,
                 step=1,
                 title=None,
                 annotation=None,
                 interval=1,
                 bgcolor=None):
        # row containing labels
        # with horizontal box layout
        row1 = QtWidgets.QWidget()
        if bgcolor:
            row1.setStyleSheet("background-color: {}".format(bgcolor.hex))
        row1_layout = QtWidgets.QHBoxLayout()
        row1_layout.setContentsMargins(12, 6, 12, 0)
        row1.setLayout(row1_layout)
        # the title label
        # if provided
        if title:
            row1_layout.addWidget(QtWidgets.QLabel(str(title)))
        # the label containing the current value
        # pushed to the right
        # and potentially with an annotation
        value_label = QtWidgets.QLabel(str(value))
        row1_layout.addStretch(1)
        row1_layout.addWidget(value_label)
        if annotation:
            row1_layout.addWidget(QtWidgets.QLabel(str(annotation)))
        # row containing slider
        row2 = QtWidgets.QWidget()
        if bgcolor:
            row2.setStyleSheet("background-color: {}".format(bgcolor.hex))
        row2_layout = QtWidgets.QHBoxLayout()
        row2_layout.setContentsMargins(12, 0, 12, 6)
        row2.setLayout(row2_layout)
        # slider
        slider = QtWidgets.QSlider()
        slider.setOrientation(QtCore.Qt.Horizontal)
        slider.setValue(value)
        slider.setMinimum(minval)
        slider.setMaximum(maxval)
        slider.setTickInterval(interval)
        slider.setSingleStep(step)
        slider.setStyleSheet(Slider.STYLE)
        row2_layout.addWidget(slider)
        # combine rows in grid
        grid = QtWidgets.QWidget()
        grid_layout = QtWidgets.QGridLayout()
        grid_layout.setSpacing(0)
        grid_layout.addWidget(row1, 0, 0)
        grid_layout.addWidget(row2, 1, 0)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid.setLayout(grid_layout)
        parent.addWidget(grid)
        # connect slider to actions
        slider.valueChanged.connect(lambda v: value_label.setText(str(v)))
        slider.valueChanged.connect(self)
        slider.valueChanged.connect(app.view.update)
        slider.sliderReleased.connect(app.view.update)
        # define attributes
        self.action = action
        self.slider = slider

    def __call__(self, *args, **kwargs):
        """Wrapper for the action associated with the slider.

        Returns
        -------
        None

        """
        return self.action(*args, **kwargs)

    @property
    def value(self):
        return self.slider.value()

    @value.setter
    def value(self, value):
        self.slider.setValue(value)



from compas.datastructures import Mesh
from compas_view2.shapes import Arrow
from .meshobject import MeshObject


class ArrowObject(MeshObject):
    """Object for displaying COMPAS arrow geometry."""

    def __init__(self, data, u=16, **kwargs):
        super().__init__(Mesh.from_shape(data, u=u), **kwargs)
        self.u = u
        self._data = data

    def update(self):
        self._mesh = Mesh.from_shape(self._data, u=self.u)
        self.init()
        super().update()

    @property
    def properties(self):
        return ["u"]

    @classmethod
    def create_default(cls) -> Arrow:
        return Arrow([0, 0, 0], [0, 0, 1])



import abc
import numpy as np
import inspect

from compas.geometry import Transformation
from compas.geometry import Translation
from compas.geometry import Rotation
from compas.geometry import Scale
from compas.geometry import decompose_matrix
from compas.geometry import identity_matrix
from compas.colors import Color
from compas.data import Data

from typing import Dict, Union


ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

DATA_OBJECT = {}


def _get_object_cls(data):
    dtype = type(data)
    cls = None

    for type_ in inspect.getmro(dtype):
        cls = DATA_OBJECT.get(type_, None)
        if cls is not None:
            break

    if cls is None:
        raise Exception('No object is registered for this data type: {}'.format(dtype))

    return cls


class Object(ABC):
    """Base object for compas_view2

    Parameters
    ----------
    data: :class:`compas.data.Data`
        A COMPAS data object.
    is_selected : bool, optional
        Whether the object is selected.
        Default to False.
    is_visible : bool, optional
        Whether to show object.
        Default to True.
    show_points : bool, optional
        Whether to show points/vertices of the object.
        Default to False.
    show_lines : bool, optional
        Whether to show lines/edges of the object.
        Default to True.
    show_faces : bool, optional
        Whether to show faces of the object.
        Default to True.
    pointcolor : Union[Color, Dict[Union[str, int], Color]], optional
        The color or the dict of colors of the points.
        Default to `compas_view2.objects.Object.default_color_points`.
    linecolor : Union[Color, Dict[Union[str, int], Color]], optional
        The color or the dict of colors of the lines.
        Default to `compas_view2.objects.Object.default_color_lines`.
    facecolor : Union[Color, Dict[Union[str, int], Color]], optional
        The color or the dict of colors the faces.
        Default to `compas_view2.objects.Object.default_color_faces`.
    linewidth : int, optional
        The line width to be drawn on screen
        Default to 1.
    pointsize : int, optional
        The point size to be drawn on screen
        Default to 10.
    opacity : float, optional
        The opacity of the object.
        Default to 1.0.
    **kwargs : dict, optional
        Additional visualization options for specific objects.

    Attributes
    ----------
    is_selected : bool
        Whether the object is selected.
    is_visible : bool
        Whether to show object.
    show_points : bool
        Whether to show points/vertices of the object.
    show_lines : bool
        Whether to show lines/edges of the object.
    show_faces : bool
        Whether to show faces of the object.
    pointcolor : :class:`compas.color.Color`
        The color of the points.
    linecolor : :class:`compas.color.Color`
        The color of the lines.
    facecolor : :class:`compas.color.Color`
        The color of the faces.
    pointcolors : Dict[Union[str, int], Color]
        The color dict of individual points.
    linecolors : Dict[Union[str, int], Color]
        The color dict of individual lines.
    facecolors : Dict[Union[str, int], Color]
        The color dict of individual faces.
    linewidth : int
        The line width to be drawn on screen.
    pointsize : int
        The point size to be drawn on screen.
    opacity : float
        The opacity of the object.
    background : bool
        Whether the object is drawn on the backgound with depth test disabled.
    bounding_box : :class:`numpy.array`, read-only
        The min and max corners of object bounding box, as a numpy array of shape (2, 3).
    bounding_box_center : :class:`numpy.array`, read-only
        The center of object bounding box, as a numpy array of shape (3,).
    matrix : :class:`numpy.array`
        The transformation matrix of the object.
    translation : :class:`numpy.array`
        The translation vector of the object.
    rotation : :class:`numpy.array`
        The euler rotation vector of the object.
    scale : :class:`numpy.array`
        The scale vector of the object.
    properties : list, read-only
        The list of object-specific properties.
    otype : class
        The data class of the object.

    Class Attributes
    ----------------
    default_color_points : :class:`compas.color.Color`
        The default color of the points with value rgb(0.2, 0.2, 0.2).
    default_color_lines : :class:`compas.color.Color`
        The default color of the lines with value rgb(0.4, 0.4, 0.4).
    default_color_faces : :class:`compas.color.Color`
        The default color of the faces with value rgb(0.8, 0.8, 0.8).

    """

    default_color_points = Color(0.2, 0.2, 0.2)
    default_color_lines = Color(0.4, 0.4, 0.4)
    default_color_faces = Color(0.8, 0.8, 0.8)

    @staticmethod
    def register(dtype, otype):
        """Register an object class to its corrensponding data type"""
        DATA_OBJECT[dtype] = otype

    @staticmethod
    def build(data, **kwargs):
        """Build an object class according to its corrensponding data type"""
        try:
            obj = _get_object_cls(data)(data, **kwargs)
        except KeyError:
            raise TypeError("Type {} is not supported by the viewer.".format(type(data)))
        return obj

    def __init__(self,
                 data: Data,
                 app=None,
                 name: str = None,
                 is_selected: bool = False,
                 is_visible: bool = True,
                 show_points: bool = False,
                 show_lines: bool = True,
                 show_faces: bool = True,
                 pointcolor: Union[Color, Dict[Union[str, int], Color]] = None,
                 linecolor: Union[Color, Dict[Union[str, int], Color]] = None,
                 facecolor: Union[Color, Dict[Union[str, int], Color]] = None,
                 linewidth: int = 1,
                 pointsize: int = 10,
                 opacity: float = 1.0):

        self._data = data
        self._app = app
        self.name = name or str(self)
        self.is_selected = is_selected
        self.is_visible = is_visible
        self.parent = None
        self._children = set()

        self.show_points = show_points
        self.show_lines = show_lines
        self.show_faces = show_faces

        if isinstance(pointcolor, dict):
            self.pointcolor = Color(*self.default_color_points)
            self.pointcolors = pointcolor
        else:
            self.pointcolor = Color(*(pointcolor or self.default_color_points))
            self.pointcolors = {}
        if isinstance(linecolor, dict):
            self.linecolor = Color(*self.default_color_lines)
            self.linecolors = linecolor
        else:
            self.linecolor = Color(*(linecolor or self.default_color_lines))
            self.linecolors = {}
        if isinstance(facecolor, dict):
            self.facecolor = Color(*self.default_color_faces)
            self.facecolors = facecolor
        else:
            self.facecolor = Color(*(facecolor or self.default_color_faces))
            self.facecolors = {}

        self.linewidth = linewidth
        self.pointsize = pointsize
        self.opacity = opacity
        self.background = False

        self._instance_color = None
        self._translation = [0., 0., 0.]
        self._rotation = [0., 0., 0.]
        self._scale = [1., 1., 1.]
        self._transformation = Transformation()
        self._matrix_buffer = None

        self._bounding_box = None
        self._bounding_box_center = None
        self._is_collection = False

    @property
    def bounding_box(self):
        return self._bounding_box

    @property
    def bounding_box_center(self):
        return self._bounding_box_center

    @property
    def otype(self):
        return DATA_OBJECT[self._data.__class__]

    @property
    def DATA_OBJECT(self):
        return DATA_OBJECT

    @abc.abstractmethod
    def init(self):
        pass

    @abc.abstractmethod
    def draw(self, shader):
        pass

    def create(self):
        pass

    @property
    def properties(self):
        return None

    @property
    def children(self):
        return self._children

    def add(self, item, **kwargs):
        if isinstance(item, Object):
            obj = item
        else:
            obj = self._app.add(item, **kwargs)
        self._children.add(obj)
        obj.parent = self

        if self._app.dock_slots['sceneform'] and self._app.view.isValid():
            self._app.dock_slots['sceneform'].update()
        return obj

    def remove(self, obj):
        obj.parent = None
        self._children.remove(obj)

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, vector):
        self._translation[0] = vector[0]
        self._translation[1] = vector[1]
        self._translation[2] = vector[2]

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, angles):
        self._rotation[0] = angles[0]
        self._rotation[1] = angles[1]
        self._rotation[2] = angles[2]

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, factors):
        self._scale[0] = factors[0]
        self._scale[1] = factors[1]
        self._scale[2] = factors[2]

    def _update_matrix(self):
        """Update the matrix from object's translation, rotation and scale"""
        if (not self.parent or self.parent._matrix_buffer is None) and (self.translation == [0, 0, 0] and self.rotation == [0, 0, 0] and self.scale == [1, 1, 1]):
            self._transformation.matrix = identity_matrix(4)
            self._matrix_buffer = None
        else:
            T1 = Translation.from_vector(self.translation)
            R1 = Rotation.from_euler_angles(self.rotation)
            S1 = Scale.from_factors(self.scale)
            M = T1 * R1 * S1
            self._transformation.matrix = M.matrix
            self._matrix_buffer = np.array(self.matrix_world).flatten()

        if self.children:
            for child in self.children:
                child._update_matrix()

    @property
    def transformation(self):
        return self._transformation

    @property
    def transformation_world(self):
        """Get the updated matrix from object's translation, rotation and scale"""
        if self.parent:
            return self.parent.transformation_world * self.transformation
        else:
            return self.transformation

    @property
    def matrix(self):
        """Get the updated matrix from object's translation, rotation and scale"""
        return self.transformation.matrix

    @property
    def matrix_world(self):
        """Get the updated matrix from object's translation, rotation and scale"""
        return self.transformation_world.matrix

    @matrix.setter
    def matrix(self, matrix):
        """Set the object's translation, rotation and scale from given matrix, and update object's matrix"""
        scale, _, rotation, tranlation, _ = decompose_matrix(matrix)
        self.translation = tranlation
        self.rotation = rotation
        self.scale = scale
        self._update_matrix()



import sys
import traceback
import time
from qtpy.QtCore import QObject
from qtpy.QtCore import QRunnable
from qtpy.QtCore import QThreadPool
from qtpy.QtCore import Slot
from qtpy.QtCore import Signal


class WorkerSignals(QObject):
    """Signals to allow worker to communicate with outside in event-driven manner
    """
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    tick = Signal()
    progress = Signal(object)


class Worker(QRunnable):
    """Worker instance that runs in a independent thread without blocking the main process
    """

    pool = QThreadPool()

    def __init__(self, fn, args=[], kwargs={}, include_self=False):
        super(Worker, self).__init__()
        self.fn = fn
        if include_self:
            args = [self] + args
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()
        self.no_signals = False

    @Slot()
    def run(self):
        """Execute the worker function, send on signals on different scenarios
        """
        if self.no_signals:
            result = self.fn(*self.args, **self.kwargs)
        else:
            try:
                result = self.fn(*self.args, **self.kwargs)
            except Exception:
                traceback.print_exc()
                exctype, value = sys.exc_info()[:2]
                self.signals.error.emit(
                    (exctype, value, traceback.format_exc()))
            else:
                # Return the result of the processing
                self.signals.result.emit(result)
            finally:
                self.signals.finished.emit()  # Done


class Ticker(QRunnable):

    pool = QThreadPool()

    def __init__(self, interval=1):
        super(Ticker, self).__init__()
        self.interval = interval
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        """Execute the worker function, send on signals on different scenarios
        """
        self.running = True
        while self.running:
            self.signals.tick.emit()
            time.sleep(self.interval)

    def stop(self):
        self.running = False



from compas.geometry import Translation
from compas.geometry import Rotation
from compas.geometry import Vector
from numpy.linalg import norm
from numpy.linalg import det
from math import atan2
from numpy import pi
from numpy import array
from numpy import asfortranarray
from numpy import dot
from numpy import float32
from compas_view2.objects import Object
from typing import List

from .matrices import perspective, ortho


class Position(Vector):

    def __init__(self, vector, on_update=None):
        super().__init__(*vector)
        self.pause_update = False
        if on_update:
            self.on_update = on_update

    def set(self, x, y, z, pause_update=False):
        pause_update = pause_update or self.pause_update
        if hasattr(self, 'on_update') and not pause_update:
            self.on_update([x, y, z])
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        if hasattr(self, 'on_update') and not self.pause_update:
            self.on_update([x, self.y, self.z])
        self._x = float(x)

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        if hasattr(self, 'on_update') and not self.pause_update:
            self.on_update([self.x, y, self.z])
        self._y = float(y)

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, z):
        if hasattr(self, 'on_update') and not self.pause_update:
            self.on_update([self.x, self.y, z])
        self._z = float(z)


class RotationEuler(Position):
    pass


class Camera:
    """Camera object for the default view.

    Parameters
    ----------
    view: :class:`compas_view2.views.View`,
        The parent view of the camera.
    fov: float, optional
        The field-of-view of the camera in degrees.
    near: float, optional
        Distance to the near clipping plane.
    far: float, optional
        Distance to the far clipping plane.
    position: list[float], optional
        The location the camera.
    target: list[float], optional
        The target location the camera is aimed at.
        Default is None, in which case the origin of the world coordinate system is used.

    Attributes
    ----------
    fov : float
        The field of view as an angler in degrees.
    near : float
        The location of the "near" clipping plane.
    far : float
        The locaiton of the "far" clipping plane.
    position : :class:`compas_view2.scene.camera.Position`
        The location the camera.
    rotation : :class:`compas_view2.scene.camera.RotationEuler`
        The euler rotation of camera.
    target : :class:`compas_view2.scene.camera.Position`
        The viewing target.
        Default is the origin of the world coordinate system.
    distance : float
        The distance from the camera standpoint to the target.
    zoom_delta : float
        Size of one zoom increment.
    rotation_delta : float
        Size of one rotation increment.
    pan_delta : float
        Size of one pan increment.

    Notes
    -----
    Under construction...

    """

    def __init__(self, view, fov=45, near=0.1, far=1000, position=None, target=None):
        self.view = view
        self.fov = fov
        self.near = near
        self.far = far
        self._position = Position([0, 0, 10], on_update=self._on_position_update)
        self._rotation = RotationEuler([0, 0, 0], on_update=self._on_rotation_update)
        self._target = Position([0, 0, 0], on_update=self._on_target_update)
        self.zoom_delta = 0.05
        self.rotation_delta = 0.01
        self.pan_delta = 0.05
        self.reset_position()
        if target:
            self.target = target
        if position:
            self.position = position

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._position.set(*position)

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, rotation):
        self._rotation.set(*rotation)

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target):
        self._target.set(*target)

    @property
    def distance(self):
        return (self.position - self.target).length

    @distance.setter
    def distance(self, distance):
        """Update the position based on the distance."""
        direction = self.position - self.target
        direction.unitize()
        new_position = self.target + direction * distance
        self.position.set(*new_position, pause_update=True)

    def _on_position_update(self, new_position):
        """Update camera rotation to keep pointing the target."""
        old_direction = array(self.position - self.target)
        new_direction = array(Vector(*new_position) - self.target)
        old_distance = norm(old_direction)
        new_distance = norm(new_direction)
        self.distance *= new_distance / old_distance

        old_direction_xy = old_direction[:2]
        new_direction_xy = new_direction[:2]
        old_direction_xy_distance = norm(old_direction_xy)
        new_direction_xy_distance = norm(new_direction_xy)

        old_direction_pitch = array([old_direction_xy_distance, old_direction[2]])
        new_direction_pitch = array([new_direction_xy_distance, new_direction[2]])
        old_direction_pitch_distance = norm(old_direction_pitch)
        new_direction_pitch_distance = norm(new_direction_pitch)

        old_direction_xy /= old_direction_xy_distance
        new_direction_xy /= new_direction_xy_distance
        old_direction_pitch /= old_direction_pitch_distance
        new_direction_pitch /= new_direction_pitch_distance

        angle_z = atan2(det([old_direction_xy, new_direction_xy]), dot(old_direction_xy, new_direction_xy))
        angle_x = -atan2(det([old_direction_pitch, new_direction_pitch]), dot(old_direction_pitch, new_direction_pitch))

        new_rotation = self.rotation + [angle_x or 0, 0, angle_z or 0]
        self.rotation.set(*new_rotation, pause_update=True)

    def _on_rotation_update(self, rotation):
        """Update camera position when rotation around target."""
        R = Rotation.from_euler_angles(rotation)
        T = Translation.from_vector([0, 0, self.distance])
        M = (R * T).matrix
        vector = [M[i][3] for i in range(3)]
        position = self.target + vector

        self.position.set(*position, pause_update=True)

    def _on_target_update(self, target):
        """Update camera position when target changes."""
        R = Rotation.from_euler_angles(self.rotation)
        T = Translation.from_vector([0, 0, self.distance])
        M = (R * T).matrix
        vector = [M[i][3] for i in range(3)]
        position = Vector(*target) + Vector(*vector)

        self.target.set(*target, pause_update=True)
        self.position.set(*position, pause_update=True)

    def reset_position(self):
        """Reset the position of the camera based current view type."""
        self.target.set(0, 0, 0)
        if self.view.current == self.view.PERSPECTIVE:
            self.rotation.set(pi/4, 0, -pi/4)
        if self.view.current == self.view.TOP:
            self.rotation.set(0, 0, 0)
        if self.view.current == self.view.FRONT:
            self.rotation.set(pi/2, 0, 0)
        if self.view.current == self.view.RIGHT:
            self.rotation.set(pi/2, 0, pi/2)

    def rotate(self, dx, dy):
        """Rotate the camera based on current mouse movement.

        Parameters
        ----------
        dx : float
            Number of rotation increments around the Z axis, with each increment the size of :attr:`Camera.rotation_delta`.
        dy: float
            Number of rotation increments around the X axis, with each increment the size of :attr:`Camera.rotation_delta`.

        Returns
        -------
        None

        Notes
        -----
        Camera rotations are only available if the current view is a perspective view (``camera.view.current == camera.view.PERSPECTIVE``).

        """
        if self.view.current == self.view.PERSPECTIVE:
            self.rotation += [-self.rotation_delta * dy, 0, -self.rotation_delta * dx]

    def pan(self, dx, dy):
        """Pan the camera based on current mouse movement.

        Parameters
        ----------
        dx : float
            Number of "pan" increments in the "X" direction of the current view, with each increment the size of :attr:`Camera.pan_delta`.
        dy : float
            Number of "pan" increments in the "Y" direction of the current view, with each increment the size of :attr:`Camera.pan_delta`.

        Returns
        -------
        None

        """
        R = Rotation.from_euler_angles(self.rotation)
        T = Translation.from_vector([-dx * self.pan_delta, dy * self.pan_delta, 0])
        M = (R * T).matrix
        vector = [M[i][3] for i in range(3)]
        self.target += vector

    def zoom(self, steps=1):
        """Zoom in or out.

        Parameters
        ----------
        steps : int
            The number of zoom increments, with each increment the zsize of :attr:`Camera.zoom_delta`.

        Returns
        -------
        None

        """
        self.distance -= steps * self.zoom_delta * self.distance

    def projection(self, width, height):
        """Compute the projection matrix corresponding to the current camera settings.

        Parameters
        ----------
        width : float
            Width of the viewer.
        height : float
            Height of the viewer.

        Returns
        -------
        4x4 np.array
            The transformation matrix as a `numpy` array in column-major order.

        Notes
        -----
        The projection matrix transforms the scene from camera coordinates to screen coordinates.

        """
        aspect = width / height
        if self.view.current == self.view.PERSPECTIVE:
            P = perspective(self.fov, aspect, self.near, self.far)
        else:
            left = -self.distance
            right = self.distance
            bottom = -self.distance / aspect
            top = self.distance / aspect
            P = ortho(left, right, bottom, top, self.near, self.far)
        return asfortranarray(P, dtype=float32)

    def viewworld(self):
        """Compute the view-world matrix corresponding to the current camera settings.

        Returns
        -------
        4x4 np.array
            The transformation matrix in column-major order.

        Notes
        -----
        The view-world matrix transforms the scene from world coordinates to camera coordinates.

        """
        T = Translation.from_vector(self.position)
        R = Rotation.from_euler_angles(self.rotation)
        W = T * R
        return asfortranarray(W.inverted(), dtype=float32)

    def zoom_extents(self, objects: List[Object] = None):

        objects = objects or self.view.objects.values()

        extents = []

        for obj in objects:
            if not obj.is_visible:
                continue

            if obj.bounding_box is None and hasattr(obj, '_update_bounding_box'):
                obj._update_bounding_box()

            if obj.bounding_box is not None:
                extents.append(obj.bounding_box)

        extents = array([obj.bounding_box for obj in objects if obj.bounding_box is not None])

        if len(extents) == 0:
            return

        extents = extents.reshape(-1, 3)
        max_corner = extents.max(axis=0)
        min_corner = extents.min(axis=0)
        center = (max_corner + min_corner) / 2
        distance = norm(max_corner - min_corner)

        self.target = center
        self.position = center + self.position
        self.distance = distance * 1.5



from .matrices import ortho  # noqa : F401
from .matrices import perspective  # noqa : F401
from .matrices import lookat  # noqa : F401

from .camera import Camera  # noqa : F401
from .mouse import Mouse  # noqa : F401



import time
from random import randint
import numpy as np

from .worker import Ticker


class Selector:
    """Selector class manages all selection operations for the viewer.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The parent application.

    Attributes
    ----------
    app : :class:`compas_view2.app.App`
        The parent application.
    mode : "single" | "multi" | "deselect"
        The selection mode.
    overwrite_mode : "single" | "multi" | "deselect"
        Used in interative selection sessions to temporarily overwrite default select mode
    types : list
        Selectable types.
    select_from : "pixel" | "box"
        The selection mechanism.
    enabled : bool
        Flag indicating to the view that an instance map should be drawn.
    wait_for_selection : bool
        Flag indicating to wait for the interatice selection to finish
    wait_for_selection_on_plane : bool
        Flag indicating to wait for the user to pick a location on plane
    snap_to_grid : bool
        Turn grid snap on or off.
    colors_to_exclude : list[tuple[int, int, int]]
        The instance colors to exclude from the selection process.
    instances : dict
        Mapping between pixel colors and scene objects.
    instance_map : np.array with shape (?,?,3)
        The painted instance map as np array
    box_select_coords : list of 4 floats
        The 2D box selection coordinates on view window: [minX, minY, maxX, maxY]
    location_on_plane :
        The selected location on plane
    selected : list of instances
        The instances that are selected

    """

    def __init__(self, app):
        self.app = app
        # Selector options
        self.mode = "single"
        self.overwrite_mode = None
        self.types = []
        self.select_from = "pixel"  # or "box"
        # Selector state flags
        self.enabled = False
        self.wait_for_selection = False
        self.wait_for_selection_on_plane = False
        self.snap_to_grid = False
        # Selector data
        self.colors_to_exclude = [(0, 0, 0,), (255, 255, 255)]
        self.instances = {}
        self.instance_map = None
        self.box_select_coords = np.zeros((4,), np.int)
        self.location_on_plane = None
        self.start_monitor_instance_map()

    # -------------------------------------------------------------------------
    # properties
    # -------------------------------------------------------------------------

    @property
    def selected(self):
        return [self.instances[key] for key in self.instances if self.instances[key].is_selected]

    # -------------------------------------------------------------------------
    # methods
    # -------------------------------------------------------------------------

    def reset(self):
        """Reset the selector state.

        Returns
        -------
        None

        """
        self.mode = "single"
        self.overwrite_mode = None
        self.types = []
        self.select_from = "pixel"
        self.enabled = False
        self.box_select_coords = np.zeros((4,), np.int)
        self.wait_for_selection = False
        self.wait_for_selection_on_plane = False
        self.location_on_plane = None
        self.snap_to_grid = False
        self.deselect()

    def start_monitor_instance_map(self):
        """This function triggers a monitor loop to watch the instance map attribute,
        Once an instance map is painted, the actual selection operation is triggered here.

        Returns
        -------
        None

        """
        def select():
            if self.instance_map is not None:
                instance_map = self.instance_map
                if self.select_from == "pixel":
                    # Pick an object from mouse pixel
                    x = self.app.view.mouse.last_pos.x()
                    y = self.app.view.mouse.last_pos.y()
                    self.select_one_from_instance_map(x, y, instance_map)
                if self.select_from == "box":
                    # Pick objects from box selection
                    self.select_all_from_instance_map(instance_map)
                    self.select_from = "pixel"
                self.app.view.update()
                self.instance_map = None

                if self.app.dock_slots["propertyform"] is not None and self.selected:
                    self.app.dock_slots["propertyform"].set_object(self.selected[0])
                if self.app.dock_slots["sceneform"] is not None:
                    self.app.dock_slots["sceneform"].select(self.selected)
                for func in self.app.on_object_selected:
                    func(self.selected)

        # Start monitor loop in a separate worker thread
        ticker = Ticker(interval=0.02)
        ticker.signals.tick.connect(select)
        Ticker.pool.start(ticker)

        self.app._app.aboutToQuit.connect(ticker.stop)

    def get_rgb_key(self):
        """
        Returns
        -------
        rgb_key
            a tuple of rgb color value in integer
        """
        while True:
            rgb_key = (randint(0, 255), randint(0, 255), randint(0, 255))
            if rgb_key not in self.instances and rgb_key not in self.colors_to_exclude:
                return rgb_key

    def add(self, obj):
        """Add an object to the list of selector instances, each object will be assigned a unique hex color key

        Returns
        -------
        rgb_key
            a rgb tuple key that represents this object
        """
        rgb_key = self.get_rgb_key()
        self.instances[rgb_key] = obj
        obj._instance_color = np.array(rgb_key)/255
        return rgb_key

    def select_one_from_instance_map(self, x, y, instance_map):
        """Select the object at given pixel location of the instance map

        Parameters
        ----------
        x : int
            x coordinate of the pixel
        y : int
            y coordinate of the pixel
        instance_map: np.array
            instance map of the current camera view

        Returns
        -------
        None
        """
        rgb_key = tuple(instance_map[y][x])
        obj = None
        if rgb_key in self.instances:
            obj = self.instances[rgb_key]
        self.select(obj)

    def select_all_from_instance_map(self, instance_map):
        """Select all the objects that appear in the instance map

        Parameters
        ----------
        instance_map: np.array
            instance map of the current camera view

        Returns
        -------
        None
        """
        unique_rgbs = np.unique(instance_map.reshape(-1, instance_map.shape[2]), axis=0)
        for rgb in unique_rgbs:
            rgb_key = tuple(rgb)
            if rgb_key in self.instances:
                obj = self.instances[rgb_key]
                self.select(obj)

    def select(self, obj=None, mode=None, types=None, update=False):
        """Perform select operation

        Parameters
        ----------
        obj : compas_view2.objects.Object
            the target object
        mode : string
            the selection mode, can be "single", "multi" or "deselect"
        types : type
            object data types to be allowed
        update : bool
            whether to trigger app update after this selection

        Returns
        -------
        None
        """
        mode = mode or self.mode
        types = types or self.types
        if mode == 'single':
            if obj:
                self.deselect()
                obj.is_selected = True
        elif mode == 'multi':
            if not obj:
                return
            if types:
                for _type in types:
                    if isinstance(obj._data, _type):
                        obj.is_selected = True
            else:
                obj.is_selected = True
        elif mode == 'deselect':
            self.deselect(obj)
        else:
            raise NotImplementedError
        if update:
            self.app.view.update()

    def deselect(self, obj=None, update=False):
        """Deselect the target object

        Parameters
        ----------
        obj : compas_view2.objects.Object
            the target object
        update : bool
            whether to trigger app update after this selection

        Returns
        -------
        None
        """
        if obj:
            obj.is_selected = False
        else:
            for key in self.instances:
                self.instances[key].is_selected = False
        if update:
            self.app.view.update()

    def start_selection(self, types=None, mode="multi", returns="data"):
        """Start an interactive selection session.

        Parameters
        ----------
        types : list of type
            the allowed types of object data
        mode : string
            the selection mode of the session, default to "multi"
        returns : string
            controls whether to return the objects or the data of objects

        Returns
        -------
        list of selected object data

        Notes
        -----
        This function has to be called inside a interactive (non-blocking) session,
        Otherwise it will freeze the main programme.
        """
        if not isinstance(types, list) and types is not None:
            types = [types]
        self.deselect(update=True)
        self.mode = self.overwrite_mode = mode
        self.types = types
        self.wait_for_selection = True
        while self.wait_for_selection:
            time.sleep(0.05)
        if returns == "data":
            selected = [obj._data for obj in self.selected]
        if returns == "object":
            selected = [obj for obj in self.selected]
        else:
            raise ValueError("must choose to return 'data' or 'object'")
        self.reset()
        return selected

    def start_selection_on_plane(self, snap_to_grid=False):
        """Start an interactive selection session to pick a location on the grid plane.

        Parameters
        ----------
        snap_to_grid : bool
            Whether to snap the location on the grid

        Returns
        -------
        List of 3 float numbers, selected location on the plane

        Notes
        -----
        This function has to be called inside a interactive (non-blocking) session,
        Otherwise it will freeze the main programme.
        """
        self.wait_for_selection_on_plane = True
        self.snap_to_grid = snap_to_grid
        while self.wait_for_selection_on_plane:
            time.sleep(0.05)
        return self.location_on_plane

    def finish_selection_on_plane(self, x, y):
        """Finish selecting location on the grid plane.

        Parameters
        ----------
        x : int
            mouse x coordinate on uv_plane_map
        y : int
            mouse y coordinate on uv_plane_map

        Returns
        -------
        None
        """
        u, v, w = self.app.selector.uv_plane_map[y, x]
        if [u, v, w] == [1, 1, 1]:
            # When clicked outside of the grid
            self.location_on_plane = None
        else:
            x_size = self.app.view.grid.x_cells * self.app.view.grid.cell_size
            y_size = self.app.view.grid.y_cells * self.app.view.grid.cell_size
            x = -x_size + 2*x_size*u
            y = -y_size + 2*y_size*v
            if self.snap_to_grid:
                x = round(x / self.app.view.grid.cell_size) * self.app.view.grid.cell_size
                y = round(y / self.app.view.grid.cell_size) * self.app.view.grid.cell_size
            self.location_on_plane = [x, y, 0]
        self.wait_for_selection_on_plane = False

    def finish_selection(self):
        """Finish the interactive selection session.
        """
        self.wait_for_selection = False
        self.wait_for_selection_on_plane = False

    def reset_box_selection(self, x, y):
        """Reset box selection start position

        Parameters
        ----------
        x : int
            current x coordinate of the mouse
        y : int
            current y coordinate of the mouse

        Returns
        -------
        None
        """
        # Set the start mouse location
        self.box_select_coords[:2] = x, y

    def perform_box_selection(self, x, y):
        """Update a box selection session.

        Parameters
        ----------
        x : int
            current x coordinate of the mouse
        y : int
            current y coordinate of the mouse

        Returns
        -------
        None
        """
        # Set the current mouse location
        box_width = abs(x - self.box_select_coords[0])
        box_height = abs(y - self.box_select_coords[1])

        if box_width > 0 and box_height > 0:
            self.select_from = "box"
            self.box_select_coords[2:] = x, y



from qtpy import QtWidgets


class Select:
    """Class representing a selection list.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The app containing the widget.
    parent : QtWidgets.QWidget
        The parent widget for the combo box.
    items : list[dict[str, Any]]
        A list of selection options, with each item defined as a dict with a particular structure.
        See Notes for more info.
    action : callable
        The action associated with the combo box.

    Attributes
    ----------
    combo : QtWidgets.QComboBox
        The combo box containing the selection options.
    action : callable
        The action associated with the index change event of the combo box.

    Notes
    -----
    Every `item` dict should have the following structure.

    .. code-block:: python

        item['text']     # str  : The text label of the item.

    """

    def __init__(self, app, parent, *, items, action):

        container = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        combo = QtWidgets.QComboBox()
        for item in items:
            combo.addItem(item['text'])
        layout.addWidget(combo)
        container.setLayout(layout)
        parent.addWidget(container)
        combo.currentIndexChanged.connect(self)
        combo.currentIndexChanged.connect(app.view.update)
        self.combo = combo
        self.action = action

    def __call__(self, *args, **kwargs):
        """Wrapper for the action associated with the combo box.

        Returns
        -------
        None

        """
        index = self.combo.currentIndex()
        text = self.combo.currentText()
        self.action(index, text)



import numpy as np

from .bufferobject import BufferObject


class CompositeObject(BufferObject):
    """Object for displaying a composition of View2 objects."""

    def __init__(self, objects, **kwargs):
        self.objects = objects
        super().__init__([obj._data for obj in objects], **kwargs)

    def _points_data(self):
        positions, colors, elements = [], [], []
        for obj in self.objects:
            if hasattr(obj, "_points_data"):
                p, c, e = obj._points_data()
                positions += p
                colors += c
                elements += e
        return positions, colors, elements

    def _lines_data(self):
        positions, colors, elements = [], [], []
        for obj in self.objects:
            if hasattr(obj, "_lines_data"):
                p, c, e = obj._lines_data()
                elements += (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
        return positions, colors, elements

    def _frontfaces_data(self):
        positions, colors, elements = [], [], []
        for obj in self.objects:
            if hasattr(obj, "_frontfaces_data"):
                p, c, e = obj._frontfaces_data()
                elements += (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
        return positions, colors, elements

    def _backfaces_data(self):
        positions, colors, elements = [], [], []
        for obj in self.objects:
            if hasattr(obj, "_backfaces_data"):
                p, c, e = obj._backfaces_data()
                elements += (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
        return positions, colors, elements



import os

__author__ = ["tom van mele"]
__copyright__ = "Block Research Group - ETH Zurich"
__license__ = "MIT License"
__email__ = "van.mele@arch.ethz.ch"
__version__ = "0.8.0"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))


os.environ['QT_MAC_WANTS_LAYER'] = '1'
os.environ['QT_API'] = 'pyside2'


try:
    from OpenGL import GL  # noqa: F401
except Exception:
    from ctypes import util
    orig_util_find_library = util.find_library

    def new_util_find_library(name):
        res = orig_util_find_library(name)
        if res:
            return res
        return '/System/Library/Frameworks/' + name + '.framework/' + name
    util.find_library = new_util_find_library


__all__ = ["HOME", "DATA", "DOCS", "TEMP", "DATA_OBJECT", "register"]



from compas.geometry import Line
from compas.geometry import Frame
from compas.datastructures import Mesh
from compas.geometry import Plane

from .meshobject import MeshObject
from .lineobject import LineObject
from .compositeobject import CompositeObject


class PlaneObject(CompositeObject):
    """Object for displaying COMPAS Plane geometry."""

    def __init__(self, data, size=1, **kwargs):
        self._data = data
        self.frame = Frame.from_plane(data)

        line = Line(
            self.frame.to_world_coordinates([0, 0, 0]),
            self.frame.to_world_coordinates([0, 0, size])
        )
        lineObject = LineObject(line, **kwargs)

        vertices = [
            self.frame.to_world_coordinates([-size, -size, 0]),
            self.frame.to_world_coordinates([size, -size, 0]),
            self.frame.to_world_coordinates([size, size, 0]),
            self.frame.to_world_coordinates([-size, size, 0])
        ]
        faces = [
            [0, 1, 2],
            [0, 2, 3]
        ]
        mesh = Mesh.from_vertices_and_faces(vertices, faces)
        meshObject = MeshObject(mesh, hide_coplanaredges=True, **kwargs)

        super().__init__([meshObject, lineObject], **kwargs)

    @classmethod
    def create_default(cls) -> Plane:
        return Plane([0, 0, 0], [0, 0, 1])



from qtpy import QtWidgets

from .form import Form


class PointForm(Form):

    def __init__(self):
        super().__init__(title='Add Point')

    def inputs(self):
        inputs = QtWidgets.QVBoxLayout()
        # inputs: X
        x_layout = QtWidgets.QHBoxLayout()
        x_label = QtWidgets.QLabel('X')
        self.x_input = QtWidgets.QDoubleSpinBox()
        self.x_input.setValue(0.0)
        x_layout.addWidget(x_label)
        x_layout.addWidget(self.x_input)
        inputs.addLayout(x_layout)
        # inputs: Y
        y_layout = QtWidgets.QHBoxLayout()
        y_label = QtWidgets.QLabel('Y')
        self.y_input = QtWidgets.QDoubleSpinBox()
        self.y_input.setValue(0.0)
        y_layout.addWidget(y_label)
        y_layout.addWidget(self.y_input)
        inputs.addLayout(y_layout)
        # inputs: Z
        z_layout = QtWidgets.QHBoxLayout()
        z_label = QtWidgets.QLabel('Z')
        self.z_input = QtWidgets.QDoubleSpinBox()
        self.z_input.setValue(0.0)
        z_layout.addWidget(z_label)
        z_layout.addWidget(self.z_input)
        inputs.addLayout(z_layout)
        return inputs

    @property
    def x(self):
        return float(self.x_input.value())

    @property
    def y(self):
        return float(self.y_input.value())

    @property
    def z(self):
        return float(self.z_input.value())



import math

from compas.geometry import Frame
from compas.geometry import Plane
from compas.geometry import Circle

from .polylineobject import PolylineObject


class CircleObject(PolylineObject):
    """Object for displaying COMPAS Circle geometry."""

    def __init__(self, circle, u=16, **kwargs):
        self.u = u
        self.calculate_circle_points(circle)
        super().__init__(circle, close=True, **kwargs)

    def calculate_circle_points(self, circle):
        frame = Frame.from_plane(circle.plane)
        circle.points = [
            frame.to_world_coordinates([
                math.cos(i*math.pi*2/self.u) * circle.radius,
                math.sin(i*math.pi*2/self.u) * circle.radius,
                0
            ])
            for i in range(self.u)
        ]

    def update(self):
        self.calculate_circle_points(self._data)
        self.init()
        super().update()

    @property
    def properties(self):
        return ["u"]

    @classmethod
    def create_default(cls) -> Circle:
        return Circle(Plane([0, 0, 0], [0, 0, 1]), 1)



from qtpy import QtWidgets

from .form import Form


class TorusForm(Form):

    def __init__(self):
        super().__init__(title='Add Torus')

    def inputs(self):
        # inputs
        inputs = QtWidgets.QVBoxLayout()
        # inputs: radius
        size_layout = QtWidgets.QHBoxLayout()
        # radius
        radius_label = QtWidgets.QLabel('Radius')
        self.radius_input = QtWidgets.QDoubleSpinBox()
        self.radius_input.setValue(1.0)
        size_layout.addWidget(radius_label)
        size_layout.addWidget(self.radius_input)
        # tube
        tube_label = QtWidgets.QLabel('Tube')
        self.tube_input = QtWidgets.QDoubleSpinBox()
        self.tube_input.setValue(1.0)
        size_layout.addWidget(tube_label)
        size_layout.addWidget(self.tube_input)
        inputs.addLayout(size_layout)
        # inputs: UV
        uv_layout = QtWidgets.QHBoxLayout()
        # U
        u_label = QtWidgets.QLabel('U')
        self.u_input = QtWidgets.QSpinBox()
        self.u_input.setValue(16)
        self.u_input.setSingleStep(4)
        uv_layout.addWidget(u_label)
        uv_layout.addWidget(self.u_input)
        # V
        v_label = QtWidgets.QLabel('V')
        self.v_input = QtWidgets.QSpinBox()
        self.v_input.setValue(16)
        self.v_input.setSingleStep(4)
        uv_layout.addWidget(v_label)
        uv_layout.addWidget(self.v_input)
        inputs.addLayout(uv_layout)
        return inputs

    @property
    def radius(self):
        return float(self.radius_input.value())

    @property
    def tube(self):
        return float(self.tube_input.value())

    @property
    def u(self):
        return int(self.u_input.value())

    @property
    def v(self):
        return int(self.v_input.value())
